@@ -2,7 +2,7 @@
  *
  *  Connection Manager
  *
- *  Copyright (C) 2007-2009  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2007-2010  Intel Corporation. All rights reserved.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License version 2 as
@@ -32,15 +32,37 @@ extern "C" {
  * @short_description: Functions for handling properties
  */
 
+#define CONNMAN_PROPERTY_ID_NAME	"Name"
+#define CONNMAN_PROPERTY_ID_TYPE	"Type"
+#define CONNMAN_PROPERTY_ID_PRIORITY	"Priority"
+#define CONNMAN_PROPERTY_ID_STRENGTH	"Strength"
+
 enum connman_property_id {
 	CONNMAN_PROPERTY_ID_INVALID = 0,
 
+	CONNMAN_PROPERTY_ID_HOSTNAME,
+	CONNMAN_PROPERTY_ID_DOMAINNAME,
+
 	CONNMAN_PROPERTY_ID_IPV4_METHOD,
 	CONNMAN_PROPERTY_ID_IPV4_ADDRESS,
 	CONNMAN_PROPERTY_ID_IPV4_NETMASK,
 	CONNMAN_PROPERTY_ID_IPV4_GATEWAY,
 	CONNMAN_PROPERTY_ID_IPV4_BROADCAST,
 	CONNMAN_PROPERTY_ID_IPV4_NAMESERVER,
+	CONNMAN_PROPERTY_ID_IPV4_TIMESERVER,
+	CONNMAN_PROPERTY_ID_IPV4_PAC,
+
+	CONNMAN_PROPERTY_ID_IPV6_METHOD,
+	CONNMAN_PROPERTY_ID_IPV6_ADDRESS,
+	CONNMAN_PROPERTY_ID_IPV6_PREFIXLEN,
+	CONNMAN_PROPERTY_ID_IPV6_GATEWAY,
+};
+
+enum connman_property_type {
+	CONNMAN_PROPERTY_TYPE_INVALID = 0,
+	CONNMAN_PROPERTY_TYPE_STRING,
+	CONNMAN_PROPERTY_TYPE_UINT8,
+	CONNMAN_PROPERTY_TYPE_BLOB,
 };
 
 struct connman_property {
@@ -48,7 +70,7 @@ struct connman_property {
 	int type;
 	int subtype;
 	void *value;
-	int size;
+	unsigned int size;
 };
 
 #ifdef __cplusplus
