@@ -41,6 +41,7 @@
 #include <connman/option.h>
 #include <connman/inet.h>
 #include <connman/dbus.h>
+#include <connman/wifi.h>
 #include <connman/log.h>
 
 #include "supplicant.h"
@@ -174,6 +175,13 @@ struct supplicant_result {
 	dbus_int32_t maxrate;
 };
 
+struct supplicant_block {
+	unsigned char *ssid;
+	char *netpath;
+	gboolean enabled;
+	int num_scans;
+};
+
 struct supplicant_task {
 	int ifindex;
 	char *ifname;
@@ -183,6 +191,8 @@ struct supplicant_task {
 	struct connman_network *pending_network;
 	char *path;
 	char *netpath;
+	gboolean hidden_found;
+	GHashTable *hidden_blocks;
 	gboolean created;
 	enum supplicant_state state;
 	gboolean scanning;
@@ -206,6 +216,16 @@ static void free_task(struct supplicant_task *task)
 	g_free(task);
 }
 
+static void remove_block(gpointer user_data)
+{
+	struct supplicant_block *block = user_data;
+
+	DBG("");
+
+	g_free(block->ssid);
+	g_free(block->netpath);
+}
+
 static struct supplicant_task *find_task_by_index(int index)
 {
 	GSList *list;
@@ -303,12 +323,266 @@ static int get_bssid(struct connman_device *device,
 	return 0;
 }
 
+static int enable_network(struct supplicant_task *task, const char *netpath,
+			  connman_bool_t enable)
+{
+	DBusMessage *message, *reply;
+	DBusError error;
+	char *enable_string;
+
+	DBG("enable %d", enable);
+
+	enable_string = enable ? "enable" : "disable";
+
+	message = dbus_message_new_method_call(SUPPLICANT_NAME, netpath,
+				SUPPLICANT_INTF ".Network", enable_string);
+	if (message == NULL)
+		return -ENOMEM;
+
+	dbus_message_set_auto_start(message, FALSE);
+
+	dbus_error_init(&error);
+
+	reply = dbus_connection_send_with_reply_and_block(connection,
+							message, -1, &error);
+	if (reply == NULL) {
+		if (dbus_error_is_set(&error) == TRUE) {
+			connman_error("%s", error.message);
+			dbus_error_free(&error);
+		} else
+			connman_error("Failed to select network");
+		dbus_message_unref(message);
+		return -EIO;
+	}
+
+	dbus_message_unref(reply);
+
+	dbus_message_unref(message);
+
+	return 0;
+}
+
+static int set_hidden_network(struct supplicant_task *task, const char *netpath,
+				const unsigned char *ssid, int ssid_len)
+{
+	DBusMessage *message, *reply;
+	DBusMessageIter array, dict;
+	DBusError error;
+	dbus_uint32_t scan_ssid = 1;
+	const char *invalid_address = "ff:ff:ff:ff:ff:ff";
+
+	message = dbus_message_new_method_call(SUPPLICANT_NAME, netpath,
+					SUPPLICANT_INTF ".Network", "set");
+	if (message == NULL)
+		return -ENOMEM;
+
+	dbus_message_set_auto_start(message, FALSE);
+
+	dbus_message_iter_init_append(message, &array);
+
+	connman_dbus_dict_open(&array, &dict);
+
+	connman_dbus_dict_append_basic(&dict, "scan_ssid",
+					 DBUS_TYPE_UINT32, &scan_ssid);
+
+	connman_dbus_dict_append_fixed_array(&dict, "ssid",
+					DBUS_TYPE_BYTE, &ssid, ssid_len);
+
+	/*
+	 * We're setting an invalid BSSID to prevent wpa_s from associating
+	 * automatically to this block once it's found.
+	 */
+	connman_dbus_dict_append_basic(&dict, "bssid",
+					DBUS_TYPE_STRING, &invalid_address);
+
+	connman_dbus_dict_close(&array, &dict);
+
+	dbus_error_init(&error);
+
+	reply = dbus_connection_send_with_reply_and_block(connection,
+							message, -1, &error);
+	if (reply == NULL) {
+		if (dbus_error_is_set(&error) == TRUE) {
+			connman_error("%s", error.message);
+			dbus_error_free(&error);
+		} else
+			connman_error("Failed to set network options");
+		dbus_message_unref(message);
+		return -EIO;
+	}
+
+	dbus_message_unref(reply);
+
+	dbus_message_unref(message);
+
+	return 0;
+}
+
+static void block_reset(gpointer key, gpointer value, gpointer user_data)
+{
+	struct supplicant_block *block = value;
+	struct supplicant_task *task = user_data;
+
+	block->num_scans = 0;
+	if (block->enabled)
+		enable_network(task, block->netpath, FALSE);
+
+	block->enabled = FALSE;
+}
+
+#define MAX_BLOCK_SCANS 2
+static void hidden_block_enable(struct supplicant_task *task)
+{
+	GHashTableIter iter;
+	gpointer key, value;
+	struct supplicant_block *block;
+
+	DBG("network %p", task->network);
+
+	if (g_hash_table_size(task->hidden_blocks) == 0)
+		return;
+
+	/*
+	 * If we're associated or associating, we no longer need to
+	 * look for hidden networks.
+	 */
+	if (task->network)
+		return;
+
+	/*
+	 * We go through the block list and:
+	 * - If we scanned it more than twice, we disable it and move
+	 *   on to the next block.
+	 * - If the next block is not enabled, we enable it, start
+	 *   the scan counter, and return. This routine will be called
+	 *   again when the next scan results are available.
+	 * - If we're done with all the blocks there, we just reset them.
+	 */
+	g_hash_table_iter_init(&iter, task->hidden_blocks);
+	while (g_hash_table_iter_next(&iter, &key, &value)) {
+		block = value;
+
+		DBG("%s num of scans %d enabled %d",
+			block->ssid, block->num_scans, block->enabled);
+
+		if (block->num_scans > MAX_BLOCK_SCANS) {
+			if (block->enabled == FALSE)
+				continue;
+
+			enable_network(task, block->netpath, FALSE);
+			block->enabled = FALSE;
+			continue;
+		}
+
+		if (block->enabled == FALSE) {
+			enable_network(task, block->netpath, TRUE);
+			block->enabled = TRUE;
+		}
+
+		block->num_scans++;
+
+		return;
+	}
+
+	g_hash_table_foreach(task->hidden_blocks, block_reset, task);
+}
+
+static int add_hidden_network(struct supplicant_task *task,
+				const unsigned char *ssid, int ssid_len)
+{
+	DBusMessage *message, *reply;
+	DBusError error;
+	const char *path;
+	struct supplicant_block *block;
+	char *netpath = NULL;
+	int ret, i;
+
+	DBG("task %p", task);
+
+	message = dbus_message_new_method_call(SUPPLICANT_NAME, task->path,
+				SUPPLICANT_INTF ".Interface", "addNetwork");
+	if (message == NULL)
+		return -ENOMEM;
+
+	dbus_message_set_auto_start(message, FALSE);
+
+	dbus_error_init(&error);
+
+	reply = dbus_connection_send_with_reply_and_block(connection,
+							message, -1, &error);
+	if (reply == NULL) {
+		if (dbus_error_is_set(&error) == TRUE) {
+			connman_error("%s", error.message);
+			dbus_error_free(&error);
+		} else
+			connman_error("Failed to add network");
+		dbus_message_unref(message);
+		return -EIO;
+	}
+
+	dbus_error_init(&error);
+
+	if (dbus_message_get_args(reply, &error, DBUS_TYPE_OBJECT_PATH, &path,
+						DBUS_TYPE_INVALID) == FALSE) {
+		if (dbus_error_is_set(&error) == TRUE) {
+			connman_error("%s", error.message);
+			dbus_error_free(&error);
+		} else
+			connman_error("Wrong arguments for network");
+		dbus_message_unref(reply);
+		return -EIO;
+	}
+
+	netpath = g_strdup(path);
+
+	ret = set_hidden_network(task, netpath, ssid, ssid_len);
+	if (ret < 0)
+		goto done;
+
+	block = g_try_new0(struct supplicant_block, 1);
+	if (block == NULL)
+		goto done;
+
+	block->ssid = g_try_malloc0(ssid_len + 1);
+	if (block->ssid == NULL) {
+		g_free(block);
+		goto done;
+	}
+
+	for (i = 0; i < ssid_len; i++) {
+		if (g_ascii_isprint(ssid[i]))
+			block->ssid[i] = ssid[i];
+		else
+			block->ssid[i] = ' ';
+	}
+
+	block->netpath = netpath;
+	block->enabled = FALSE;
+	block->num_scans = 0;
+
+	DBG("path %s ssid %s", block->netpath, block->ssid);
+
+	g_hash_table_replace(task->hidden_blocks, block->ssid, block);
+
+	return 0;
+done:
+	g_free(netpath);
+
+	dbus_message_unref(reply);
+
+	dbus_message_unref(message);
+
+	return ret;
+}
+
 static void add_interface_reply(DBusPendingCall *call, void *user_data)
 {
 	struct supplicant_task *task = user_data;
 	DBusMessage *reply;
 	DBusError error;
 	const char *path;
+	char **hex_ssids, *hex_ssid;
+	int i;
 
 	DBG("task %p", task);
 
@@ -340,6 +614,33 @@ static void add_interface_reply(DBusPendingCall *call, void *user_data)
 
 	dbus_pending_call_unref(call);
 
+	hex_ssids = connman_wifi_load_ssid();
+	if (hex_ssids == NULL)
+		return;
+
+	for (i = 0; hex_ssids[i]; i++) {
+		unsigned char *ssid;
+		unsigned int j, k = 0, hex;
+		size_t hex_ssid_len;
+
+		hex_ssid = hex_ssids[i];
+		hex_ssid_len = strlen(hex_ssid);
+
+		ssid = g_try_malloc0(hex_ssid_len / 2 + 1);
+		if (ssid == NULL)
+			break;
+
+		for (j = 0, k = 0; j < hex_ssid_len; j += 2) {
+			sscanf(hex_ssid + j, "%02x", &hex);
+			ssid[k++] = hex;
+		}
+
+		if (add_hidden_network(task, ssid, hex_ssid_len / 2) < 0)
+			break;
+	}
+
+	g_strfreev(hex_ssids);
+
 	return;
 
 failed:
@@ -774,29 +1075,44 @@ static int set_network_tls(struct connman_network *network,
 	const char *private_key_password;
 
 	/*
-	 * For TLS, we at least need a key, the client cert,
-	 * and a passhprase.
-	 * Server cert is optional.
+	 * For TLS, we at least need:
+	 *              The client certificate
+	 *              The client private key file
+	 *              The client private key file password
+	 *
+	 * The Authority certificate is optional.
 	 */
 	client_cert = connman_network_get_string(network,
 						"WiFi.ClientCertFile");
-	if (client_cert == NULL)
+	if (client_cert == NULL) {
+		connman_error("Error in TLS authentication: "
+			      "a ClientCertFile must be defined\n");
 		return -EINVAL;
+	}
 
 	private_key = connman_network_get_string(network,
 						"WiFi.PrivateKeyFile");
-	if (private_key == NULL)
+	if (private_key == NULL) {
+		connman_error("Error in TLS authentication: "
+			      "a PrivateKeyFile must be defined\n");
 		return -EINVAL;
+	}
 
 	private_key_password = connman_network_get_string(network,
 						"WiFi.PrivateKeyPassphrase");
-	if (private_key_password == NULL)
+	if (private_key_password == NULL) {
+		connman_error("Error in TLS authentication: "
+			      "a PrivateKeyPassphrase must be defined\n");
 		return -EINVAL;
+	}
 
 	ca_cert = connman_network_get_string(network, "WiFi.CACertFile");
 	if (ca_cert)
 		connman_dbus_dict_append_basic(dict, "ca_cert",
 						DBUS_TYPE_STRING, &ca_cert);
+	else
+		connman_info("No CACertFile has been provided "
+			     "to do the TLS authentication\n");
 
 	DBG("client cert %s private key %s", client_cert, private_key);
 
@@ -818,19 +1134,35 @@ static int set_network_peap(struct connman_network *network,
 	char *phase2_auth;
 
 	/*
-	 * For PEAP, we at least need the sever cert, a 2nd
-	 * phase authentication and a passhprase.
-	 * Client cert is optional although strongly required
-	 * When setting the client cert, we then need a private
-	 * key as well.
+	 * For PEAP/TTLS, we at least need
+	 *              The authority certificate
+	 *              The 2nd phase authentication method
+	 *              The 2nd phase passphrase
+	 *
+	 * The Client certificate is optional although strongly required
+	 * When setting it, we need in addition
+	 *              The Client private key file
+	 *              The Client private key file password
 	 */
-	ca_cert = connman_network_get_string(network, "WiFi.CACertFile");
-	if (ca_cert == NULL)
+	if (passphrase == NULL) {
+		connman_error("Error in PEAP/TTLS authentication: "
+			      "a phase2 passphrase must be defined\n");
 		return -EINVAL;
+	}
+
+	ca_cert = connman_network_get_string(network, "WiFi.CACertFile");
+	if (ca_cert == NULL) {
+		connman_error("Error in PEAP/TTLS authentication: "
+			      "CACertFile must be defined\n");
+		return -EINVAL;
+	}
 
 	phase2 = connman_network_get_string(network, "WiFi.Phase2");
-	if (phase2 == NULL)
+	if (phase2 == NULL) {
+		connman_error("Error in PEAP/TTLS authentication: "
+			      "Phase2 must be defined\n");
 		return -EINVAL;
+	}
 
 	DBG("CA cert %s phase2 auth %s", ca_cert, phase2);
 
@@ -841,14 +1173,22 @@ static int set_network_peap(struct connman_network *network,
 
 		private_key = connman_network_get_string(network,
 							"WiFi.PrivateKeyFile");
-		if (private_key == NULL)
+		if (private_key == NULL) {
+			connman_error("Error in PEAP/TTLS authentication: "
+				      "with ClientCertFile, "
+				      "PrivateKeyFile must be defined\n");
 			return -EINVAL;
+		}
 
 		private_key_password =
 			connman_network_get_string(network,
 						"WiFi.PrivateKeyPassphrase");
-		if (private_key_password == NULL)
+		if (private_key_password == NULL) {
+			connman_error("Error in PEAP/TTLS authentication: "
+				      "with ClientCertFile, "
+				      "PrivateKeyPassphrase must be defined\n");
 			return -EINVAL;
+		}
 
 		connman_dbus_dict_append_basic(dict, "client_cert",
 						DBUS_TYPE_STRING, &client_cert);
@@ -861,7 +1201,9 @@ static int set_network_peap(struct connman_network *network,
 							&private_key_password);
 
 		DBG("client cert %s private key %s", client_cert, private_key);
-	}
+	} else
+		connman_info("No client certificate has been provided "
+			     "to do the PEAP/TTLS authentication\n");
 
 	phase2_auth = g_strdup_printf("\"auth=%s\"", phase2);
 
@@ -961,7 +1303,8 @@ static int set_network(struct supplicant_task *task,
 				dbus_message_unref(message);
 				return err;
 			}
-		} else if (g_strcmp0(eap, "peap") == 0) {
+		} else if (g_strcmp0(eap, "peap") == 0 ||
+				   g_strcmp0(eap, "ttls") == 0) {
 			int err;
 
 			err = set_network_peap(network, &dict, passphrase);
@@ -1021,6 +1364,17 @@ static int set_network(struct supplicant_task *task,
 						"wep_key0", DBUS_TYPE_BYTE,
 							&key, size / 2);
 				free(key);
+			} else if (size == 5 || size == 13) {
+				unsigned char *key = malloc(13);
+				int i;
+				if (key == NULL)
+					size = 0;
+				for (i = 0; i < size; i++)
+					key[i] = (unsigned char) passphrase[i];
+				connman_dbus_dict_append_fixed_array(&dict,
+						"wep_key0", DBUS_TYPE_BYTE,
+								&key, size);
+				free(key);
 			} else
 				connman_dbus_dict_append_basic(&dict,
 						"wep_key0", DBUS_TYPE_STRING,
@@ -1426,6 +1780,7 @@ static void properties_reply(DBusPendingCall *call, void *user_data)
 {
 	struct supplicant_task *task = user_data;
 	struct supplicant_result result;
+	struct supplicant_block *block;
 	struct connman_network *network;
 	DBusMessage *reply;
 	DBusMessageIter array, dict;
@@ -1517,6 +1872,17 @@ static void properties_reply(DBusPendingCall *call, void *user_data)
 	if (result.path[0] == '\0')
 		goto done;
 
+	if (result.name) {
+		block = g_hash_table_lookup(task->hidden_blocks, result.name);
+		if (block) {
+			enable_network(task, block->netpath, FALSE);
+			g_hash_table_remove(task->hidden_blocks, block->ssid);
+		}
+	}
+
+	if (result.ssid == NULL)
+		task->hidden_found = TRUE;
+
 	if (result.frequency > 0 && result.frequency < 14)
 		result.frequency = 2407 + (5 * result.frequency);
 	else if (result.frequency == 14)
@@ -1618,8 +1984,17 @@ static void get_properties(struct supplicant_task *task)
 	char *path;
 
 	path = g_slist_nth_data(task->scan_results, 0);
-	if (path == NULL)
+	if (path == NULL) {
+		if (task->hidden_found == TRUE) {
+			/*
+			 * We're done with regular scanning, let's enable
+			 * the missing network blocks if there are hidden
+			 * SSIDs around.
+			 */
+			hidden_block_enable(task);
+		}
 		goto noscan;
+	}
 
 	message = dbus_message_new_method_call(SUPPLICANT_NAME, path,
 						SUPPLICANT_INTF ".BSSID",
@@ -1702,6 +2077,8 @@ static void scan_results_reply(DBusPendingCall *call, void *user_data)
 		task->scan_results = g_slist_append(task->scan_results, path);
 	}
 
+	task->hidden_found = FALSE;
+
 	g_strfreev(results);
 
 	dbus_message_unref(reply);
@@ -1792,6 +2169,8 @@ static int task_connect(struct supplicant_task *task)
 	unsigned int ssid_len;
 	int err;
 
+	g_hash_table_foreach(task->hidden_blocks, block_reset, task);
+
 	connman_inet_ifup(task->ifindex);
 
 	address = connman_network_get_string(task->network, "Address");
@@ -1845,6 +2224,9 @@ static void scanning(struct supplicant_task *task, DBusMessage *msg)
 
 	connman_info("%s scanning %s", task->ifname,
 				scanning == TRUE ? "started" : "finished");
+
+	if (scanning == TRUE)
+		task->scanning = TRUE;
 }
 
 static void state_change(struct supplicant_task *task, DBusMessage *msg)
@@ -2025,7 +2407,8 @@ int supplicant_start(struct connman_device *device)
 	task->state = WPA_INVALID;
 	task->disconnecting = FALSE;
 	task->pending_network = NULL;
-
+	task->hidden_blocks = g_hash_table_new_full(g_str_hash, g_str_equal,
+							NULL, remove_block);
 	task_list = g_slist_append(task_list, task);
 
 	return create_interface(task);
@@ -2052,6 +2435,7 @@ int supplicant_stop(struct connman_device *device)
 	g_free(task->range);
 
 	task_list = g_slist_remove(task_list, task);
+	g_hash_table_destroy(task->hidden_blocks);
 
 	if (task->scan_call != NULL) {
 		dbus_pending_call_cancel(task->scan_call);
@@ -2161,6 +2545,25 @@ int supplicant_disconnect(struct connman_network *network)
 	return 0;
 }
 
+void supplicant_remove_network(struct connman_network *network)
+{
+	struct supplicant_task *task;
+	int index;
+
+	DBG("network %p", network);
+
+	index = connman_network_get_index(network);
+
+	task = find_task_by_index(index);
+	if (task == NULL)
+		return;
+
+	if (task->network != network)
+		return;
+
+	remove_network(task);
+}
+
 static void supplicant_activate(DBusConnection *conn)
 {
 	DBusMessage *message;
