@@ -172,6 +172,12 @@ int main(int argc, char *argv[])
 			perror("Failed to create storage directory");
 	}
 
+	if (mkdir(STORAGEDIR "/stats", S_IRUSR | S_IWUSR | S_IXUSR |
+				S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH) < 0) {
+		if (errno != EEXIST)
+			perror("Failed to create statistics directory");
+	}
+
 	old_umask = umask(077);
 
 	main_loop = g_main_loop_new(NULL, FALSE);
@@ -219,10 +225,13 @@ int main(int argc, char *argv[])
 	__connman_element_init(option_device, option_nodevice);
 
 	__connman_agent_init();
+	__connman_tethering_init();
 	__connman_counter_init();
+	__connman_ondemand_init();
 	__connman_manager_init(option_compat);
 	__connman_profile_init();
 	__connman_config_init();
+	__connman_stats_init();
 
 	__connman_resolver_init();
 	__connman_ipconfig_init();
@@ -260,11 +269,14 @@ int main(int argc, char *argv[])
 	__connman_ipconfig_cleanup();
 	__connman_resolver_cleanup();
 
+	__connman_stats_cleanup();
 	__connman_config_cleanup();
 	__connman_profile_cleanup();
 	__connman_manager_cleanup();
+	__connman_ondemand_cleanup();
 	__connman_counter_cleanup();
 	__connman_agent_cleanup();
+	__connman_tethering_cleanup();
 
 	__connman_element_cleanup();
 	__connman_storage_cleanup();
